@@ -6,6 +6,7 @@
 -include_lib("diameter_3gpp_ts29_272.hrl").
 -include_lib("osmo_gsup/include/gsup_protocol.hrl").
 
+-define(DIA_VENDOR_3GPP, 10415).
 
 %% diameter callbacks
 -export([peer_up/3, peer_down/3, pick_peer/4, prepare_request/3, prepare_retransmit/3,
@@ -123,7 +124,7 @@ req_resynchronization_info(_) ->
 gsup_pdp2dia(GsupPdpInfo) ->
 	#'PDP-Context'{'PDP-Type' = maps:get(pdp_type, GsupPdpInfo, ?PDP_TYPE_DEFAULT),
 		       'Context-Identifier' = maps:get(pdp_context_id, GsupPdpInfo),
-		       'Service-Selection' = maps:get(access_point_name, GsupPdpInfo),
+		       'Service-Selection' = decode_apn:decode_apn(maps:get(access_point_name, GsupPdpInfo)),
 		       'QoS-Subscribed' = maps:get(quality_of_service, GsupPdpInfo, ?PDP_QOS_DEFAULT)
 		      }.
 
@@ -149,7 +150,7 @@ gsup_pdp2dia_apn(GsupPdpInfo) ->
 					      'Max-Requested-Bandwidth-DL' = 100000000},
 			     % The default APN Configuration shall not contain the Wildcard APN (see 3GPP TS
 			     % 23.003 [3], clause 9.2); the default APN shall always contain an explicit APN
-			     'Service-Selection' = "internet"%maps:get(access_point_name, GsupPdpInfo)
+			     'Service-Selection' = decode_apn:decode_apn(maps:get(access_point_name, GsupPdpInfo))
 			    }.
 
 % transient (only in Experimental-Result-Code)
@@ -157,7 +158,10 @@ gsup_pdp2dia_apn(GsupPdpInfo) ->
 -define(DIAMETER_ERROR_CAMEL_SUBSCRIPTION_PRESENT,	4182).
 % permanent (only in Experimental-Result-Code)
 -define(DIAMETER_ERROR_USER_UNKNOWN,			5001).
+-define(DIAMETER_AUTHORIZATION_REJECTED,		5003).
 -define(DIAMETER_ERROR_ROAMING_NOT_ALLOWED,		5004).
+-define(DIAMETER_MISSING_AVP,				5005).
+-define(DIAMETER_UNABLE_TO_COMPLY,			5012).
 -define(DIAMETER_ERROR_UNKNOWN_EPS_SUBSCRIPTION,	5420).
 -define(DIAMETER_ERROR_RAT_NOT_ALLOWED,			5421).
 -define(DIAMETER_ERROR_EQUIPMENT_UNKNOWN,		5422).
@@ -165,23 +169,36 @@ gsup_pdp2dia_apn(GsupPdpInfo) ->
 
 % 10.5.5.14
 -define(GMM_CAUSE_IMSI_UNKNOWN,		16#02).
--define(GMM_CAUSE_PLMN_NOTALLOWED,	16#0b).
+-define(GMM_CAUSE_ILLEGAL_MS,		16#03).
 -define(GMM_CAUSE_GPRS_NOTALLOWED,	16#07).
--define(GMM_CAUSE_INV_MAND_INFO,	16#60).
+-define(GMM_CAUSE_PLMN_NOTALLOWED,	16#0b).
+-define(GMM_CAUSE_LA_NOTALLOWED,	16#0c).
+-define(GMM_CAUSE_ROAMING_NOTALLOWED,	16#0d).
+-define(GMM_CAUSE_NO_SUIT_CELL_IN_LA,	16#0f).
 -define(GMM_CAUSE_NET_FAIL,		16#11).
-% TODO: more values
+-define(GMM_CAUSE_CONGESTION,		16#16).
+-define(GMM_CAUSE_GSM_AUTH_UNACCEPT,	16#17).
+-define(GMM_CAUSE_INV_MAND_INFO,	16#60).
+-define(GMM_CAUSE_PROTO_ERR_UNSPEC,	16#6f).
 
--define(EXP_RES(Foo),	#'Experimental-Result'{'Vendor-Id'=fixme, 'Experimental-Result-Code'=Foo}).
+-define(EXP_RES(Exp),	#'Experimental-Result'{'Vendor-Id'=?DIA_VENDOR_3GPP, 'Experimental-Result-Code'=Exp}).
 
+%% see 29.272 Annex A/B
 -type empty_or_intl() :: [] | [integer()].
 -spec gsup_cause2dia(integer()) -> {empty_or_intl(), empty_or_intl()}.
 gsup_cause2dia(?GMM_CAUSE_IMSI_UNKNOWN) -> {[], [?EXP_RES(?DIAMETER_ERROR_USER_UNKNOWN)]};
-gsup_cause2dia(?GMM_CAUSE_PLMN_NOTALLOWED) -> {[], [?DIAMETER_ERROR_ROAMING_NOT_ALLOWED]};
-gsup_cause2dia(?GMM_CAUSE_GPRS_NOTALLOWED) -> {[], [?DIAMETER_ERROR_RAT_NOT_ALLOWED]};
-%gsup_cause2dia(?GMM_CAUSE_INV_MAND_INFO) ->
-%gsup_cause2dia(?GMM_CAUSE_NET_FAIL) -> 
-% TODO: more values
-gsup_cause2dia(_) -> {fixme, []}.
+gsup_cause2dia(?GMM_CAUSE_ILLEGAL_MS) -> {[], [?EXP_RES(?DIAMETER_ERROR_USER_UNKNOWN)]};
+gsup_cause2dia(?GMM_CAUSE_PLMN_NOTALLOWED) -> {[], [?EXP_RES(?DIAMETER_ERROR_ROAMING_NOT_ALLOWED)]};
+gsup_cause2dia(?GMM_CAUSE_GPRS_NOTALLOWED) -> {[], [?EXP_RES(?DIAMETER_ERROR_UNKNOWN_EPS_SUBSCRIPTION)]};
+
+gsup_cause2dia(?GMM_CAUSE_LA_NOTALLOWED) -> {[?DIAMETER_AUTHORIZATION_REJECTED], []};
+gsup_cause2dia(?GMM_CAUSE_ROAMING_NOTALLOWED) -> {[], [?EXP_RES(?DIAMETER_ERROR_ROAMING_NOT_ALLOWED)]};
+gsup_cause2dia(?GMM_CAUSE_NO_SUIT_CELL_IN_LA) -> {[], [?EXP_RES(?DIAMETER_ERROR_UNKNOWN_EPS_SUBSCRIPTION)]};
+gsup_cause2dia(?GMM_CAUSE_NET_FAIL) -> {[?DIAMETER_UNABLE_TO_COMPLY], []};
+gsup_cause2dia(?GMM_CAUSE_CONGESTION) -> {[?DIAMETER_UNABLE_TO_COMPLY], []};
+gsup_cause2dia(?GMM_CAUSE_INV_MAND_INFO) -> {[?DIAMETER_MISSING_AVP], []};
+gsup_cause2dia(?GMM_CAUSE_PROTO_ERR_UNSPEC) -> {[?DIAMETER_UNABLE_TO_COMPLY], []};
+gsup_cause2dia(_) -> {[?DIAMETER_UNABLE_TO_COMPLY], []}.
 
 % get the value for a tiven key in Map1. If not found, try same key in Map2. If not found, return Default
 -spec twomap_get(atom(), map(), map(), any()) -> any().
@@ -223,9 +240,11 @@ handle_request(#diameter_packet{msg = Req, errors = []}, _SvcName, {_, Caps}) wh
 			Resp = #'AIA'{'Session-Id'=SessionId, 'Origin-Host'=OH, 'Origin-Realm'=OR,
 				      'Result-Code'=2001, 'Auth-Session-State'=1,
 				      'Authentication-Info'=AuthInfo};
-		#{message_type := send_auth_info_err} ->
+		#{message_type := send_auth_info_err, cause:=Cause} ->
+			{Res, ExpRes} = gsup_cause2dia(Cause),
 			Resp = #'AIA'{'Session-Id'=SessionId, 'Origin-Host'=OH, 'Origin-Realm'=OR,
-				      'Result-Code'=?DIAMETER_ERROR_USER_UNKNOWN,
+				      'Result-Code'=Res,
+				      'Experimental-Result'=ExpRes,
 				      'Auth-Session-State'=1};
 		timeout ->
 			Resp = #'AIA'{'Session-Id'=SessionId, 'Origin-Host'=OH, 'Origin-Realm'=OR,
